import concurrent
from concurrent.futures import Future, ThreadPoolExecutor, ALL_COMPLETED
from dataclasses import dataclass
from typing import Callable, List, Tuple, Deque

from dragoneye.utils.app_logger import logger


@dataclass
class ThreadedFunctionData:
    callable: Callable
    args: Tuple
    error_msg: str
    timeout_msg: str = None


def execute_parallel_functions_in_threads(tasks_groups: Deque[List[ThreadedFunctionData]], max_workers, timeout=None)\
        -> List[Tuple[Future, str, str]]:
    """
    This function takes a Queue (tasks_groups: Deque[List[ThreadedFunctionData]])
    of lists, when each queued item represents a list of functions that should be run in parallel.
    """
    tasks_responses: List[Tuple[Future, str, str]] = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:

        for tasks_group in tasks_groups:
            for task in tasks_group:
                tasks_responses.append((
                    executor.submit(task.callable, *task.args),
                    task.error_msg, task.timeout_msg))

            concurrent.futures.wait([response[0] for response in tasks_responses], timeout=timeout, return_when=ALL_COMPLETED)

        timeout_tasks = [response for response in tasks_responses if response[0].running()]
        for timeout_task in timeout_tasks:
            logger.exception(timeout_task[2])
        failed_tasks = [response for response in tasks_responses if response[0].exception()]
        for failed_task in failed_tasks:
            logger.exception(failed_task[1], exc_info=failed_task[0].exception())

    return tasks_responses



from dragoneye.cloud_scanner.aws.aws_scan_settings import AwsCloudScanSettings
from dragoneye.cloud_scanner.azure.azure_scan_settings import AzureCloudScanSettings
from dragoneye.cloud_scanner.gcp.gcp_scan_settings import GcpCloudScanSettings
from dragoneye.cloud_scanner.aws.aws_scanner import AwsScanner
from dragoneye.cloud_scanner.azure.azure_scanner import AzureScanner
from dragoneye.cloud_scanner.gcp.gcp_scanner import GcpScanner
from dragoneye.cloud_scanner.aws.aws_session_factory import AwsSessionFactory
from dragoneye.cloud_scanner.azure.azure_authorizer import AzureAuthorizer
from dragoneye.cloud_scanner.gcp.gcp_credentials_factory import GcpCredentialsFactory
from dragoneye.utils.app_logger import logger, add_file_handler



import os
from typing import List, Optional

from dragoneye.cloud_scanner.base_cloud_scanner import CloudScanSettings, CloudProvider


class AwsCloudScanSettings(CloudScanSettings):
    def __init__(self,
                 commands_path: str,
                 account_name: str,
                 regions_filter: List[str] = None,
                 max_attempts: int = 10,
                 max_pool_connections: int = 50,
                 command_timeout: int = 600,
                 output_path: str = os.getcwd(),
                 should_clean_before_scan: bool = True,
                 default_region: Optional[str] = None):
        """
        The settings that the AwsScanner uses for aws scanning.

            :param commands_path: The path of a YAML file that describes the scan commands to be used.
            :param account_name: A name for the scan results.
            :param regions_filter: A list of regions to scan resources on.
            :param max_attempts: The amount of times that the boto client will attempt to call AWS's api before giving up.
            :param max_pool_connections: The maximum number of connections to keep in a connection pool.
            :param command_timeout: The timeout period for a single scan command.
            :param output_path: The directory where results will be saved. Defaults to current working directory.
            :param should_clean_before_scan: A flag that determines if prior results of this specific account (identified by account_name)
                should be deleted before scanning.
            :param default_region: The region to be used for making requests for universal services.
                If not specified, the session's default region will be used.
        """
        super().__init__(CloudProvider.AWS, account_name, should_clean_before_scan, output_path, commands_path)
        self.regions_filter: str = ','.join(regions_filter) if regions_filter else ''
        self.max_attempts: int = max_attempts
        self.max_pool_connections: int = max_pool_connections
        self.command_timeout: int = command_timeout
        self.default_region: Optional[str] = default_region



import json
import os
from abc import abstractmethod
from enum import Enum
from queue import Queue
from dragoneye.utils.app_logger import logger
from dragoneye.utils.misc_utils import load_yaml


class CloudProvider(str, Enum):
    AWS = 'aws'
    AZURE = 'azure'
    GCP = 'gcp'


class CloudScanSettings:
    def __init__(self,
                 cloud_provider: CloudProvider,
                 account_name: str,
                 should_clean_before_scan: bool,
                 output_path: str,
                 commands_path: str):
        self.cloud_provider: CloudProvider = cloud_provider
        self.account_name: str = account_name
        self.clean: bool = should_clean_before_scan
        self.output_path: str = output_path
        self.commands_path: str = commands_path


class BaseCloudScanner:
    def __init__(self, settings: CloudScanSettings):
        self.account_data_dir: str = None
        self.summary: Queue = Queue()
        self.settings: CloudScanSettings = settings

    @abstractmethod
    def scan(self) -> str:
        pass

    @staticmethod
    def _write_failures_report(directory, failures):
        with open(os.path.join(directory, 'failures-report.json'), 'w+') as failures_report:
            failures_report.write(json.dumps(failures, default=str))

    def _print_summary(self):
        logger.info("--------------------------------------------------------------------")
        failures = []
        for call_summary in self.summary.queue:
            if 'error' in call_summary or 'exception' in call_summary:
                failures.append(call_summary)

        logger.info("Summary: {} APIs called. {} errors".format(len(self.summary.queue), len(failures)))
        if len(failures) > 0:
            logger.warning("Failures:")
            for call_summary in failures:
                logger.warning(f"  {self._parse_error(call_summary)}")

        self._write_failures_report(os.path.join(self.account_data_dir, '..'), failures)

    @staticmethod
    def _is_dynamic_parameter(parameter: dict):
        value = parameter.get('Value', parameter.get('Values'))
        return isinstance(value, str) and '|' in value

    @staticmethod
    @abstractmethod
    def _parse_error(call_summary: dict) -> str:
        pass

    def _get_scan_commands(self):
        scan_commands = load_yaml(self.settings.commands_path)
        dependent_commands = []
        independent_commands = []
        for command in scan_commands:
            parameters = command.get('Parameters', [])
            if any(self._is_dynamic_parameter(parameter) for parameter in parameters):
                dependent_commands.append(command)
            else:
                independent_commands.append(command)

        return dependent_commands, independent_commands



import collections
import copy
import logging
import os.path
import os
import re
import json
import time
from functools import lru_cache
from typing import List, Dict, Optional, Deque

import urllib.parse
from botocore.exceptions import ClientError, EndpointConnectionError
from botocore.config import Config

from dragoneye.cloud_scanner.aws.aws_scan_settings import AwsCloudScanSettings
from dragoneye.config import config
from dragoneye.utils.boto_backoff import rate_limiter
from dragoneye.cloud_scanner.base_cloud_scanner import BaseCloudScanner
from dragoneye.utils.app_logger import logger
from dragoneye.utils.misc_utils import get_dynamic_values_from_files, custom_serializer, make_directory, init_directory, snakecase, \
    elapsed_time
from dragoneye.utils.threading_utils import execute_parallel_functions_in_threads, ThreadedFunctionData

MAX_RETRIES = 3


class AwsScanner(BaseCloudScanner):

    def __init__(self, session, settings: AwsCloudScanSettings):
        super().__init__(settings)
        self.session = session
        self.settings = settings
        # Services that will only be queried in the default region
        # TODO: Identify these from boto
        self.universal_services = [
            "iam",
            "route53",
            "route53domains",
            "s3",
            "cloudfront",
            "organizations",
        ]

        self.default_region = settings.default_region or self.session.region_name
        if self.default_region is None:
            raise ValueError('Default region cannot be empty. '
                             'You must specify the default region or set the AWS_DEFAULT_REGION environment variable')
        self.handler_config = Config(retries={'max_attempts': self.settings.max_attempts, 'mode': 'standard'},
                                     max_pool_connections=self.settings.max_pool_connections)
        logging.getLogger("botocore").setLevel(logging.WARN)

    @elapsed_time('Scanning AWS live environment took {} seconds')
    def scan(self) -> str:
        self.account_data_dir = init_directory(self.settings.output_path, self.settings.account_name, self.settings.clean)
        region_dict_list = self._create_regions_file_structure()
        dependent_commands, independent_commands = self._get_scan_commands()

        tasks: List[ThreadedFunctionData] = []

        for region in region_dict_list:
            tasks.append(ThreadedFunctionData(
                self._scan_region_data,
                (region, dependent_commands, independent_commands),
                'An unknown exception has occurred'
            ))

        deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()
        deque_tasks.append(tasks)
        execute_parallel_functions_in_threads(deque_tasks, len(region_dict_list))

        self._print_summary()

        return os.path.abspath(os.path.join(self.account_data_dir, '..'))

    def _create_regions_file_structure(self):
        region_list = self._get_region_list()

        with open(f"{self.account_data_dir}/describe-regions.json", "w+") as file:
            file.write(json.dumps(region_list, indent=4, sort_keys=True))

        logger.info("* Creating directory for each region name")
        region_dict_list: List[dict] = region_list["Regions"]

        for region in region_dict_list:
            make_directory(os.path.join(self.account_data_dir, region.get("RegionName", "Unknown")))

        return region_dict_list

    def _get_region_list(self):
        regions_filter = None
        if len(self.settings.regions_filter) > 0:
            regions_filter = self.settings.regions_filter.lower().split(",")
            # Force include of default region -- seems to be required
            if self.default_region not in regions_filter:
                regions_filter.append(self.default_region)

        logger.info("* Getting region names")
        ec2 = self.session.client("ec2", region_name=self.default_region)
        region_list = ec2.describe_regions()

        if regions_filter is not None:
            filtered_regions = [r for r in region_list["Regions"] if r["RegionName"] in regions_filter]
            region_list["Regions"] = filtered_regions

        return region_list

    @staticmethod
    def _get_identifier_from_parameter(parameter):
        if isinstance(parameter, list):
            identifier = parameter[0]
        else:
            identifier = parameter

        return identifier

    @staticmethod
    def _get_filename_from_parameter(parameter):
        if isinstance(parameter, list):
            if len(parameter) > 1:
                filename = parameter[1]
            elif isinstance(parameter[0], list):
                # For elbv2:describe-tags we need ResourceArns as a list like `[Arn]`
                # the yaml file specifies `[[.LoadBalancerArn]]` because just doing
                # `[.LoadBalancerArn]` presents other issues, so this extracts out the inner, inner value.
                # Similar issue for elb:describe-tags
                filename = parameter[0][0]
            else:
                filename = parameter[0]
        else:
            filename = parameter

        return urllib.parse.quote_plus(filename)

    def _get_and_save_data(self, output_file, handler, method_to_call, parameters, checks, region):
        """
        Calls the AWS API function and downloads the data

        check: Value to check and repeat the call if it fails
        summary: Keeps tracks of failures
        """
        # TODO: Decorate this with rate limiters from
        # https://github.com/Netflix-Skunkworks/cloudaux/blob/master/cloudaux/aws/decorators.py
        if os.path.isfile(output_file):
            # Data already scanned, so skip
            logger.warning("Response already present at {}".format(output_file))
            return

        call_summary = {
            "service": handler.meta.service_model.service_name,
            "action": method_to_call,
            "parameters": parameters,
            "region": region
        }

        params_string = '' if not parameters else ', '.join(f'{k}={v}' for k, v in parameters.items())
        function_msg = f'{call_summary["service"]}.{call_summary["action"]}({params_string})'
        logger.info(f'Invoking {function_msg}')
        data = AwsScanner._get_data(output_file, handler, method_to_call, parameters, checks, call_summary)
        AwsScanner._remove_unused_values(data)
        AwsScanner._save_results_to_file(output_file, data)

        logger.info(f'Results from {function_msg} were saved to {output_file}')
        self.summary.put_nowait(call_summary)

    @staticmethod
    def _get_data(output_file, handler, method_to_call, parameters, checks, call_summary):
        data = None
        try:
            for retries in range(MAX_RETRIES):
                data = AwsScanner._call_boto_function(output_file, handler, method_to_call, parameters)
                if not checks or AwsScanner._is_data_passing_check(data, checks):
                    break
                elif retries == MAX_RETRIES - 1:
                    raise Exception(
                        "One of the following checks has repeatedly failed: {}".format(
                            ', '.join(f'{check["Name"]}={check["Value"]}' for check in checks)
                        )
                    )
                else:
                    logger.info("  Sleeping and retrying")
                    time.sleep(3)

        except ClientError as ex:
            if "NoSuchBucketPolicy" in str(ex):
                # This error occurs when you try to get the bucket policy for a bucket that has no bucket policy, so this can be ignored.
                logger.warning("  - No bucket policy")
            elif "NoSuchPublicAccessBlockConfiguration" in str(ex):
                # This error occurs when you try to get the account Public Access Block policy for an account that has none, so this can be ignored.
                logger.warning("  - No public access block set")
            elif (
                    "ServerSideEncryptionConfigurationNotFoundError" in str(ex)
                    and call_summary["service"] == "s3"
                    and call_summary["action"] == "get_bucket_encryption"
            ):
                logger.warning("  - No encryption set")
            elif (
                    "NoSuchEntity" in str(ex)
                    and call_summary["action"] == "get_account_password_policy"
            ):
                logger.warning("  - No password policy set")
            elif (
                    "AccessDeniedException" in str(ex)
                    and call_summary["service"] == "organizations"
                    and call_summary["action"] == "list_accounts"
            ):
                logger.warning("  - Denied, which likely means this is not the organization root")
            elif (
                    "RepositoryPolicyNotFoundException" in str(ex)
                    and call_summary["service"] == "ecr"
                    and call_summary["action"] == "get_repository_policy"
            ):
                logger.warning("  - No policy exists")
            elif (
                    "ResourceNotFoundException" in str(ex)
                    and call_summary["service"] == "lambda"
                    and call_summary["action"] == "get_policy"
            ):
                logger.warning("  - No policy exists")
            elif (
                    "AccessDeniedException" in str(ex)
                    and call_summary["service"] == "kms"
                    and call_summary["action"] == "list_key_policies"
            ):
                logger.warning("  - Denied, which should mean this KMS has restricted access")
            elif (
                    "AccessDeniedException" in str(ex)
                    and call_summary["service"] == "kms"
                    and call_summary["action"] == "list_grants"
            ):
                logger.warning("  - Denied, which should mean this KMS has restricted access")
            elif (
                    "AccessDeniedException" in str(ex)
                    and call_summary["service"] == "kms"
                    and call_summary["action"] == "get_key_policy"
            ):
                logger.warning("  - Denied, which should mean this KMS has restricted access")
            elif (
                    "AccessDeniedException" in str(ex)
                    and call_summary["service"] == "kms"
                    and call_summary["action"] == "get_key_rotation_status"
            ):
                logger.warning("  - Denied, which should mean this KMS has restricted access")
            elif "AWSOrganizationsNotInUseException" in str(ex):
                logger.warning(' - Your account is not a member of an organization.')
            elif (
                    "EntityNotFoundException" in str(ex)
                    and call_summary["service"] == "glue"
                    and call_summary["action"] == "get_resource_policy"
            ):
                logger.warning(f' - Glue policy does not exist on region {call_summary["region"]}')
            elif (
                    "NoSuchEntity" in str(ex)
            ):
                logger.warning(f"  - {str(ex)}")
            elif (
                    "NoSuchAccessPointPolicy" in str(ex)
            ):
                logger.warning(f"  - {str(ex)}")
            elif (
                    "PolicyNotFound" in str(ex)
            ):
                logger.warning(f"  - {str(ex)}")
            else:
                logger.warning(f"ClientError {ex}")
                call_summary["exception"] = ex
        except EndpointConnectionError as ex:
            logger.warning("EndpointConnectionError: {}".format(ex))
            call_summary["exception"] = ex
        except Exception as ex:
            if "Parameter validation failed" in str(ex):
                logger.warning(f"  - {str(ex)}")
            else:
                logger.warning("Exception: {}".format(ex))
                call_summary["exception"] = ex

        return data

    @staticmethod
    @rate_limiter()
    def _call_boto_function(output_file, handler, method_to_call, parameters):
        data = {}
        if handler.can_paginate(method_to_call):
            paginator = handler.get_paginator(method_to_call)
            page_iterator = paginator.paginate(**parameters)
            for response in page_iterator:
                if not data:
                    data = response

                else:
                    logger.info("  ...paginating {}".format(output_file))
                    for k in data:
                        if isinstance(data[k], list):
                            data[k].extend(response[k])
        else:
            function = getattr(handler, method_to_call)
            data = function(**parameters)

        return data

    @staticmethod
    def _is_data_passing_check(data: dict, checks: Optional[dict]) -> bool:
        if checks:
            for check in checks:
                if data[check["Name"]] == check["Value"]:
                    pass
                else:
                    return False
        return True

    @staticmethod
    def _remove_unused_values(data: dict) -> None:
        if data is not None:
            data.pop("ResponseMetadata", None)
            data.pop("Marker", None)
            data.pop("IsTruncated", None)

    @staticmethod
    def _save_results_to_file(output_file: str, data: Optional[Dict]) -> None:
        if data is not None:
            with open(output_file, "w+") as file:
                file.write(
                    json.dumps(data, indent=4, sort_keys=True, default=custom_serializer)
                )

    def _run_scan_commands(self, region, runner):
        region = copy.deepcopy(region)
        runner = copy.deepcopy(runner)
        region_name = region["RegionName"]

        if runner['Service'] == 'configservice':
            runner['Service'] = 'config'  # This is due to service name change between API (configservice) and python SDK (config)

        client_region = region_name
        if runner['Service'] == 'globalaccelerator':
            # globalaccelerator only has api endpoint in us-west-2
            if region_name == self.default_region:
                client_region = 'us-west-2'
            else:
                return
        else:
            if not self._should_run_command_on_region(runner, region):
                return

        handler = self.session.client(
            runner["Service"], region_name=client_region,
            config=self.handler_config
        )

        filepath = os.path.join(self.account_data_dir, region_name, f'{runner["Service"]}-{runner["Request"]}')

        method_to_call = snakecase(runner["Request"])
        parameter_keys = set()
        param_groups = self._get_parameter_group(runner, self.account_data_dir, region, parameter_keys)

        tasks: List[ThreadedFunctionData] = []

        if runner.get("Parameters"):
            make_directory(filepath)
            for param_group in param_groups:
                if set(param_group.keys()) != parameter_keys:
                    continue
                unparsed_file_name = '_'.join([f'{k}-{v}' if not isinstance(v, list) else k for k, v in param_group.items()])
                file_name = urllib.parse.quote_plus(unparsed_file_name)
                output_file = f"{filepath}/{file_name}.json"
                tasks.append(ThreadedFunctionData(
                    AwsScanner._get_and_save_data,
                    (self,
                     output_file,
                     handler,
                     method_to_call,
                     param_group,
                     runner.get("Check", None),
                     region_name),
                    'exception on command {}'.format(runner),
                    'timeout on command {}'.format(runner)))
        else:
            output_file = filepath + ".json"
            tasks.append(ThreadedFunctionData(
                AwsScanner._get_and_save_data,
                (self,
                 output_file,
                 handler,
                 method_to_call,
                 {},
                 runner.get("Check", None),
                 region_name), 'exception on command {}'.format(runner), 'timeout on command {}'.format(runner)))

        deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()
        deque_tasks.append(tasks)
        execute_parallel_functions_in_threads(deque_tasks, config.get('MAX_WORKERS'), self.settings.command_timeout)

    def _scan_region_data(self, region: dict, dependent_commands: List[dict], independent_commands: List[dict]):
        non_dependable_tasks: List[ThreadedFunctionData] = []
        dependable_tasks: List[ThreadedFunctionData] = []
        deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()

        for independent_command in independent_commands:
            non_dependable_tasks.append(ThreadedFunctionData(
                self._run_scan_commands,
                (region, independent_command),
                'exception on command {}'.format(independent_command)))

        deque_tasks.append(non_dependable_tasks)

        for dependent_command in dependent_commands:
            dependable_tasks.append(ThreadedFunctionData(
                self._run_scan_commands,
                (region, dependent_command),
                'exception on command {}'.format(dependent_command)))

        for dependable_task in dependable_tasks:
            deque_tasks.append([dependable_task])
        execute_parallel_functions_in_threads(deque_tasks, config.get('MAX_WORKERS'))

    @staticmethod
    def _get_call_parameters(call_parameters: dict, parameters_def: list) -> List[dict]:
        group_param = {}
        for parameter_def in parameters_def:
            group_param[parameter_def['Name']] = parameter_def.get('Group', False)

        params = []
        keys = list(call_parameters.keys())

        if len(call_parameters) == 1:
            for key in keys:
                for value in call_parameters[key]:
                    params.append({key: value})

        elif len(call_parameters) == 2:
            for value1 in call_parameters[keys[0]]:
                if group_param[keys[1]]:
                    params.append({keys[0]: value1, keys[1]: call_parameters[keys[1]]})
                else:
                    for value2 in call_parameters[keys[1]]:
                        params.append({keys[0]: value1, keys[1]: value2})

        return params

    @staticmethod
    def _fill_simple_params(param_groups, name, value, parameter: dict):
        if not param_groups:
            param_groups = [{name: value}]
            return param_groups
        else:
            additional_param_group: List[dict] = []
            for param_group in param_groups:
                if "Values" in parameter:
                    for val in value:
                        param: dict = param_group.copy()
                        param[name] = val
                        additional_param_group.append(param)
                else:
                    param_group[name] = value
            if additional_param_group:
                param_groups.clear()
                param_groups.extend(additional_param_group)
            return param_groups

    @staticmethod
    def _fill_dynamic_params(param_groups: List[dict],
                             name: str,
                             value: str,
                             group: bool,
                             account_dir: str,
                             region: Dict[str, str]) -> List[dict]:
        depends_on_keys = re.findall(r'{{([^|]*)}}', value) or []
        region_account_dir = os.path.join(account_dir, region['RegionName'])
        if not param_groups and depends_on_keys:
            return param_groups
        if not param_groups and not depends_on_keys:
            values = get_dynamic_values_from_files(value, region_account_dir)
            if group:
                param_groups.append({name: values})
                return param_groups
            else:
                for val in values:
                    param_groups.append({name: val})
                return param_groups

        result_param_groups = []
        for param_group in param_groups:
            real_value = value
            for key in depends_on_keys:
                real_value = real_value.replace(f'{{{{{key}}}}}', param_group[key])
            dynamic_params_list: list = get_dynamic_values_from_files(real_value, region_account_dir)
            if group:
                param_group[name] = dynamic_params_list
                result_param_groups.append(param_group)
            else:
                for cached_value in dynamic_params_list:
                    clone_param_group = copy.deepcopy(param_group)
                    clone_param_group[name] = cached_value
                    result_param_groups.append(clone_param_group)
        return result_param_groups

    def _should_run_command_on_region(self, runner: dict, region_dict: dict) -> bool:
        service = runner["Service"]
        region_name = region_dict["RegionName"]

        if service in self.universal_services:
            if region_name != self.default_region:
                return False
        elif service != 'eks' and region_name not in self._get_available_regions(service):
            logger.info("Skipping region {}, as {} does not exist there".format(region_name, service))
            return False
        return True

    def _get_parameter_group(self, runner, account_dir, region, parameter_keys: set):
        param_groups = []
        for parameter in runner.get("Parameters", []):
            name = parameter["Name"]
            value = parameter.get("Value", parameter.get("Values"))
            parameter_keys.add(name)
            if not self._is_dynamic_parameter(parameter):
                param_groups = self._fill_simple_params(param_groups, name, value, parameter)
            else:
                group = parameter.get("Group", False)
                param_groups = self._fill_dynamic_params(param_groups, name, value, group, account_dir, region)
        return param_groups

    @lru_cache(maxsize=None)
    def _get_available_regions(self, service: str):
        return self.session.get_available_regions(service)

    @staticmethod
    def _parse_error(call_summary: dict) -> str:
        return "  {}.{}({}): {}".format(
            call_summary["service"],
            call_summary["action"],
            call_summary["parameters"],
            call_summary["exception"],
        )



import functools
import glob
import json
import logging
import os
from datetime import datetime
from functools import lru_cache
from shutil import rmtree
from typing import List

import backoff
import pyjq
import requests
import yaml

from dragoneye.utils.app_logger import logger


def elapsed_time(message=None):
    def _elapsed_time(function):
        @functools.wraps(function)
        def elapsed_wrapper(*arguments):
            start = datetime.now()
            return_val = function(*arguments)
            end = datetime.now()
            _message = message or 'The function ' + f'`{function.__name__}`' + 'took {} seconds'
            logger.info(_message.format((end - start).total_seconds()))
            return return_val
        return elapsed_wrapper
    return _elapsed_time


def make_directory(path) -> None:
    try:
        os.mkdir(path)
    except OSError:
        # Already exists
        pass


def init_directory(base_path: str, account_name: str, clean: bool) -> str:
    account_data_dir = f"{base_path}/account-data/{account_name}"
    if clean and os.path.exists(account_data_dir):
        rmtree(account_data_dir)

    make_directory(base_path)
    make_directory(f"{base_path}/account-data")
    make_directory(account_data_dir)
    return os.path.abspath(account_data_dir)


def load_yaml(file_path: str) -> List[dict]:
    with open(file_path, "r") as file:
        return yaml.safe_load(file)


@backoff.on_exception(backoff.expo, requests.RequestException, 3, 600)
def invoke_get_request(url: str, headers: dict, on_success=None, on_backoff=None, on_giveup=None):
    logging.getLogger('backoff').disabled = True
    on_predicate = backoff.on_predicate(backoff.expo, lambda response: response.status_code != 200,
                                        max_tries=3,
                                        max_time=600,
                                        on_success=on_success,
                                        on_backoff=on_backoff,
                                        on_giveup=on_giveup)
    func = on_predicate(lambda _url, _headers: requests.get(url=_url, headers=_headers))
    return func(url, headers)


@lru_cache(maxsize=None)
def get_dynamic_values_from_files(value: str, directory: str) -> list:
    if '|' not in value:
        return [value]

    parameter_file = value.split("|")[0]
    parameter_file = "{}/{}".format(
        directory, parameter_file
    ).strip()

    # Get array if a globbing pattern is used (ex. "*.json")
    parameter_files = glob.glob(parameter_file)
    parameters = []
    for parameter_file in parameter_files:
        if not os.path.isfile(parameter_file):
            continue

        with open(parameter_file, "r") as file:
            parameter_values = json.load(file)
            pyjq_parse_string = "|".join(value.split("|")[1:])
            for parameter in pyjq.all(pyjq_parse_string, parameter_values):
                if isinstance(parameter, list):
                    parameters.extend(parameter)
                else:
                    parameters.append(parameter)
    return parameters


def custom_serializer(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        return obj.decode()
    raise TypeError("Unknown type")


def snakecase(string):
    return string.replace("-", "_")



import os
import re
from typing import Any, Optional


def validate_uuid(value: str, error_message='Invalid uuid format'):
    _validate_regex(value, r'^[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}$', error_message)


def validate_path(value: str, error_message='Path does not exist'):
    if not os.path.exists(value):
        raise ValueError(error_message)


def validate_not_empty(value: Optional[Any], error_message='Value is None or empty'):
    if not value:
        raise ValueError(error_message)


def _validate_regex(value: str,
                    regex: str,
                    error_message: str):

    result = re.match(regex, str(value))
    if not result:
        raise ValueError(error_message)



import os

from dragoneye.cloud_scanner.base_cloud_scanner import CloudScanSettings, CloudProvider


class AzureCloudScanSettings(CloudScanSettings):
    def __init__(self,
                 commands_path: str,
                 subscription_id: str,
                 account_name: str,
                 output_path: str = os.getcwd(),
                 should_clean_before_scan: bool = True
                 ):
        """
        The settings that the AzureScanner uses for azure scanning.
        :param commands_path: The path of a YAML file that describes the scan commands to be used.
        :param subscription_id: The subscription id you wish to scan.
        :param account_name: A name for the scan results.
        :param output_path: The directory where results will be saved. Defaults to current working directory.
        :param should_clean_before_scan: A flag that determines if prior results of this specific account (identified by account_name)
            should be deleted before scanning.
        """
        super().__init__(cloud_provider=CloudProvider.AZURE,
                         account_name=account_name,
                         should_clean_before_scan=should_clean_before_scan,
                         output_path=output_path,
                         commands_path=commands_path)
        self.subscription_id = subscription_id



import sys

from setuptools import setup, find_packages
from dragoneye.version import __version__

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read().splitlines()

project_name = 'dragoneye'
with open("README.md", "r") as fh:
    long_description = fh.read()
version = __version__

if "--version" in sys.argv:
    index = sys.argv.index('--version')
    version = sys.argv[index + 1]
    sys.argv.remove("--version")
    sys.argv.remove(version)

setup(
    name=project_name,
    description='Multi-cloud data scan tool',
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version,
    author='Indeni',
    author_email='engineering@indeni.com',
    url='https://github.com/indeni/dragoneye',
    packages=find_packages(include=[project_name, f'{project_name}.*']),
    package_data={'': ['config.yaml']},
    include_package_data=True,
    keywords=['cloud', 'aws', 'azure', 'scan'],
    install_requires=requirements,
    entry_points={
        'console_scripts': [f'{project_name}=dragoneye.scan:safe_cli_entry_point']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)
